"""
All the intermediary syntax.
We can several kinds of models can be translated to this syntax.
"""

from __future__ import annotations

import operator
import re
from abc import ABC, abstractmethod
from typing import Any, ClassVar

from .cst import FONT_TAGS, ROW_TAGS, TABLE


class Drawable(ABC):
    """Abstract class to represent all the objects which are drawable."""

    RE: ClassVar[re.Pattern[str]]

    def to_markdown(self) -> str:
        """Transforms the intermediary object to it's syntax in the er markup."""
        raise NotImplementedError()

    def to_dot(self) -> str:
        """Transforms the intermediary object to it's syntax in the dot format."""
        raise NotImplementedError()

    def __eq__(self, other) -> bool:
        return self.__dict__ == other.__dict__

    @staticmethod
    @abstractmethod
    def make_from_match(match: re.Match) -> Drawable:
        """Used in the parsing of files. Transforms a regex match to a Drawable object."""

    def __str__(self) -> str:
        return self.to_markdown()


class Column(Drawable):
    """Represents a Column in the intermediaty syntax"""

    RE = re.compile(
        '(?P<primary>\*?)(?P<name>[^\s]+)\s*(\{label:\s*"(?P<label>[^"]+)"\})?'
    )

    @staticmethod
    def make_from_match(match: re.Match) -> Column:
        return Column(
            name=match.group("name"),
            type=match.group("label"),
            is_key="*" in match.group("primary"),
            is_null=not "*" in match.group("primary"),
        )

    def __init__(self, name: str, type=None, is_key: bool = False, is_null=None):
        """
        :param name: (str) Name of the column
        :param type:
        :param is_key:
        :param is_null:
        :return:
        """
        self.name = name
        self.type = type
        self.is_key = is_key
        if is_null == None:
            self.is_null = not is_key
        else:
            self.is_null = is_null

    def __lt__(self, other: Column) -> bool:
        if self.is_key > other.is_key:
            return True
        elif self.is_key < other.is_key:
            return False
        else:
            return self.name < other.name

    @property
    def key_symbol(self) -> str:
        return "*" if self.is_key else ""

    def to_markdown(self) -> str:
        return '    {}{} {{label:"{}"}}'.format(self.key_symbol, self.name, self.type)

    def to_mermaid(self) -> str:
        return " {}{} {}{}".format(
            self.key_symbol,
            self.type.replace("(", "<").replace(")", ">"),
            self.name,
            " NOT NULL" if not self.is_null else "",
        )

    def to_dot(self) -> str:
        base = ROW_TAGS.format(
            ' ALIGN="LEFT"', "{key_opening}{col_name}{key_closing}{type}{null}"
        )
        return base.format(
            key_opening="<u>" if self.is_key else "",
            key_closing="</u>" if self.is_key else "",
            col_name=FONT_TAGS.format(self.name),
            type=FONT_TAGS.format(" [{}]").format(self.type)
            if self.type is not None
            else "",
            null=" NOT NULL" if not self.is_null else "",
        )


class Relation(Drawable):
    """Represents a Relation in the intermediaty syntax"""

    RE = re.compile(
        "(?P<left_name>[^\s]+)\s*(?P<left_cardinality>[*?+1])--(?P<right_cardinality>[*?+1])\s*(?P<right_name>[^\s]+)"
    )  # noqa: E501
    cardinalities = {"*": "0..N", "?": "{0,1}", "+": "1..N", "1": "1", "": None}
    cardinalities_mermaid = {
        "*": "0..n",
        "?": "0..1",
        "+": "1..n",
    }

    @staticmethod
    def make_from_match(match: re.Match) -> Relation:
        return Relation(
            right_col=match.group("right_name"),
            left_col=match.group("left_name"),
            right_cardinality=match.group("right_cardinality"),
            left_cardinality=match.group("left_cardinality"),
        )

    def __init__(
        self, right_col, left_col, right_cardinality=None, left_cardinality=None
    ):
        if (
            right_cardinality not in self.cardinalities.keys()
            or left_cardinality not in self.cardinalities.keys()
        ):
            raise ValueError(
                'Cardinality should be in {}"'.format(self.cardinalities.keys())
            )
        self.right_col = right_col
        self.left_col = left_col
        self.right_cardinality = right_cardinality
        self.left_cardinality = left_cardinality

    def to_markdown(self) -> str:
        return "{} {}--{} {}".format(
            self.left_col,
            self.left_cardinality,
            self.right_cardinality,
            self.right_col,
        )

    def to_mermaid(self) -> str:
        normalized = (
            Relation.cardinalities_mermaid.get(k, k)
            for k in (
                self.left_col,
                self.left_cardinality,
                self.right_cardinality,
                self.right_col,
            )
        )
        return '{} "{}" -- "{}" {}'.format(*normalized)

    def graphviz_cardinalities(self, card) -> str:
        if card == "":
            return ""
        return "label=<<FONT>{}</FONT>>".format(self.cardinalities[card])

    def to_dot(self) -> str:
        if self.right_cardinality == self.left_cardinality == "":
            return ""
        cards = []
        if self.left_cardinality != "":
            cards.append("tail" + self.graphviz_cardinalities(self.left_cardinality))
        if self.right_cardinality != "":
            cards.append("head" + self.graphviz_cardinalities(self.right_cardinality))
        return '"{}" -- "{}" [{}];'.format(
            self.left_col, self.right_col, ",".join(cards)
        )

    def __eq__(self, other: Any) -> bool:
        if super().__eq__(other):
            return True
        other_inversed = Relation(
            right_col=other.left_col,
            left_col=other.right_col,
            right_cardinality=other.left_cardinality,
            left_cardinality=other.right_cardinality,
        )
        return other_inversed.__dict__ == self.__dict__


class Table(Drawable):
    """Represents a Table in the intermediaty syntax"""

    RE = re.compile("\[(?P<name>[^]]+)\]")

    def __init__(self, name: str, columns: list[Column]) -> None:
        self.name = name
        self.columns = columns

    @staticmethod
    def make_from_match(match: re.Match) -> Table:
        return Table(name=match.group("name"), columns=[])

    @property
    def header_markdown(self) -> str:
        return "[{}]".format(self.name)

    def to_markdown(self) -> str:
        return (
            self.header_markdown
            + "\n"
            + "\n".join(c.to_markdown() for c in self.columns)
        )

    def to_mermaid(self) -> str:
        columns = [c.to_mermaid() for c in self.columns]
        return (
            "class {}{{\n  ".format(self.name)
            + "\n  ".format(self.name).join(columns)  # type:ignore
            + "\n}"
        )

    @property
    def columns_sorted(self):
        return sorted(self.columns, key=operator.attrgetter("name"))

    @property
    def header_dot(self) -> str:
        return ROW_TAGS.format("", '<B><FONT POINT-SIZE="16">{}</FONT></B>').format(
            self.name
        )

    def to_dot(self) -> str:
        body = "".join(c.to_dot() for c in self.columns)
        return TABLE.format(self.name, self.header_dot, body)

    def __str__(self) -> str:
        return self.header_markdown

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Table):
            return False
        if other.name != self.name:
            return False

        if self.columns_sorted != other.columns_sorted:
            return False
        return True



from __future__ import annotations

import argparse
import base64
import copy
import sys

from pygraphviz.agraph import AGraph
from sqlalchemy.engine.url import make_url
from sqlalchemy.exc import ArgumentError

from .cst import GRAPH_BEGINNING
from .helpers import check_args
from .parser import (
    ParsingException,
    line_iterator_to_intermediary,
    markdown_file_to_intermediary,
)
from .sqla import (
    database_to_intermediary,
    declarative_to_intermediary,
    metadata_to_intermediary,
)

if sys.version_info >= (3, 8):
    from importlib.metadata import version
else:
    # importlib.metadata not available for python 3.7
    from importlib_metadata import version

__version__ = version(__package__)


def cli() -> None:
    """Entry point for the application script"""
    parser = get_argparser()

    args = parser.parse_args()
    check_args(args)
    if args.v:
        print("eralchemy2 version {}.".format(__version__))
        exit(0)
    render_er(
        args.i,
        args.o,
        include_tables=args.include_tables,
        include_columns=args.include_columns,
        exclude_tables=args.exclude_tables,
        exclude_columns=args.exclude_columns,
        schema=args.s,
    )


def get_argparser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(prog="eralchemy2")
    parser.add_argument("-i", nargs="?", help="Database URI to process.")
    parser.add_argument("-o", nargs="?", help="Name of the file to write.")
    parser.add_argument("-s", nargs="?", help="Name of the schema.")
    parser.add_argument(
        "--exclude-tables", "-x", nargs="+", help="Name of tables not to be displayed."
    )
    parser.add_argument(
        "--exclude-columns",
        nargs="+",
        help="Name of columns not to be displayed (for all tables).",
    )
    parser.add_argument(
        "--include-tables", nargs="+", help="Name of tables to be displayed alone."
    )
    parser.add_argument(
        "--include-columns",
        nargs="+",
        help="Name of columns to be displayed alone (for all tables).",
    )
    parser.add_argument("-v", help="Prints version number.", action="store_true")
    return parser


def intermediary_to_markdown(tables, relationships, output):
    """Saves the intermediary representation to markdown."""
    er_markup = _intermediary_to_markdown(tables, relationships)
    with open(output, "w") as file_out:
        file_out.write(er_markup)


def intermediary_to_mermaid(tables, relationships, output):
    """Saves the intermediary representation to markdown."""
    markup = _intermediary_to_mermaid(tables, relationships)
    md_markup = "<!--\n\n{}\n\n-->\n".format(markup)
    markup_b64 = base64.urlsafe_b64encode(markup.encode("utf8")).decode("ascii")
    md_markup += "![](https://mermaid.ink/img/{})\n".format(markup_b64)
    with open(output, "w") as file_out:
        file_out.write(md_markup)


def intermediary_to_dot(tables, relationships, output):
    """Save the intermediary representation to dot format."""
    dot_file = _intermediary_to_dot(tables, relationships)
    with open(output, "w") as file_out:
        file_out.write(dot_file)


def intermediary_to_schema(tables, relationships, output):
    """Transforms and save the intermediary representation to the file chosen."""
    dot_file = _intermediary_to_dot(tables, relationships)
    graph = AGraph()
    graph = graph.from_string(dot_file)
    extension = output.split(".")[-1]
    graph.draw(path=output, prog="dot", format=extension)


def _intermediary_to_markdown(tables, relationships):
    """Returns the er markup source in a string."""
    t = "\n".join(t.to_markdown() for t in tables)
    r = "\n".join(r.to_markdown() for r in relationships)
    return "{}\n{}".format(t, r)


def _intermediary_to_mermaid(tables, relationships):
    """Returns the er markup source in a string."""
    t = "\n".join(t.to_mermaid() for t in tables)
    r = "\n".join(r.to_mermaid() for r in relationships)
    return "classDiagram\n{}\n{}".format(t, r)


def _intermediary_to_dot(tables, relationships):
    """Returns the dot source representing the database in a string."""
    t = "\n".join(t.to_dot() for t in tables)
    r = "\n".join(r.to_dot() for r in relationships)
    return "{}\n{}\n{}\n}}".format(GRAPH_BEGINNING, t, r)


# Routes from the class name to the function transforming this class in
# the intermediary representation.
switch_input_class_to_method = {
    "MetaData": metadata_to_intermediary,
    "DeclarativeMeta": declarative_to_intermediary,
    # For compatibility with Flask-SQLAlchemy
    "_BoundDeclarativeMeta": declarative_to_intermediary,
    # Renamed in Flask-SQLAlchemy 2.3
    "DefaultMeta": declarative_to_intermediary,
}

# Routes from the mode to the method to transform the intermediary
#  representation to the desired output.
switch_output_mode_auto = {
    "er": intermediary_to_markdown,
    "mermaid": intermediary_to_mermaid,
    "graph": intermediary_to_schema,
    "dot": intermediary_to_dot,
}

# Routes from the file extension to the method to transform
# the intermediary representation to the desired output.
switch_output_mode = {
    "er": intermediary_to_markdown,
    "md": intermediary_to_mermaid,
    "dot": intermediary_to_dot,
}


def all_to_intermediary(filename_or_input, schema=None):
    """Dispatch the filename_or_input to the different function to produce the intermediary syntax.
    All the supported classes names are in `swich_input_class_to_method`.
    The input can also be a list of strings in markdown format or a filename finishing by '.er' containing markdown
    format.
    """
    # Try to convert from the name of the class
    input_class_name = filename_or_input.__class__.__name__
    try:
        this_to_intermediary = switch_input_class_to_method[input_class_name]
        tables, relationships = this_to_intermediary(filename_or_input)
        return tables, relationships
    except KeyError:
        pass

    # try to read markdown file.
    if isinstance(filename_or_input, str):
        if filename_or_input.split(".")[-1] == "er":
            return markdown_file_to_intermediary(filename_or_input)

    # try to read a markdown in a string
    if not isinstance(filename_or_input, str):
        if all(isinstance(e, str) for e in filename_or_input):
            return line_iterator_to_intermediary(filename_or_input)

    # try to read DB URI might raise ArgumentError.
    try:
        make_url(filename_or_input)
        return database_to_intermediary(filename_or_input, schema=schema)
    except ArgumentError:
        raise ValueError("Cannot process filename_or_input {}".format(input_class_name))


def get_output_mode(output, mode):
    """
    From the output name and the mode returns a the function that will transform the intermediary
    representation to the output.
    """
    if mode != "auto":
        try:
            return switch_output_mode_auto[mode]
        except KeyError:
            raise ValueError('Mode "{}" is not supported.')

    extension = output.split(".")[-1]
    try:
        return switch_output_mode[extension]
    except KeyError:
        return intermediary_to_schema


def filter_resources(
    tables,
    relationships,
    include_tables=None,
    include_columns=None,
    exclude_tables=None,
    exclude_columns=None,
):
    """
    Include the following:
        1. Tables and relationships with tables present in the include_tables (lst of str, tables names)
        2. Columns (of whichever table) present in the include_columns (lst of str, columns names)
    Exclude the following:
        1. Tables and relationships with tables present in the exclude_tables (lst of str, tables names)
        2. Columns (of whichever table) present in the exclude_columns (lst of str, columns names)
    Disclosure note:
        All relationships are taken into consideration before ignoring columns.
        In other words, if one excludes primary or foreign keys, it will still keep the relations display amongst tables
    """
    _tables = copy.deepcopy(tables)
    _relationships = copy.deepcopy(relationships)

    include_tables = include_tables or [t.name for t in _tables]
    include_columns = include_columns or [c.name for t in _tables for c in t.columns]
    exclude_tables = exclude_tables or list()
    exclude_columns = exclude_columns or list()

    _tables = [
        t for t in _tables if t.name not in exclude_tables and t.name in include_tables
    ]
    _relationships = [
        r
        for r in _relationships
        if r.right_col not in exclude_tables
        and r.left_col not in exclude_tables
        and r.right_col in include_tables
        and r.left_col in include_tables
    ]

    for t in _tables:
        t.columns = sorted(
            [
                c
                for c in t.columns
                if c.name not in exclude_columns and c.name in include_columns
            ]
        )

    return _tables, _relationships


def render_er(
    input,
    output,
    mode="auto",
    include_tables=None,
    include_columns=None,
    exclude_tables=None,
    exclude_columns=None,
    schema=None,
):
    """
    Transform the metadata into a representation.
    :param input: Possible inputs are instances of:
        MetaData: SQLAlchemy Metadata
        DeclarativeMeta: SQLAlchemy declarative Base
    :param output: name of the file to output the
    :param mode: str in list:
        'er': writes to a file the markup to generate an ER style diagram.
        'graph': writes the image of the ER diagram.
        'mermaid': writes to a file the markup to generate an Mermaid-JS style diagram
        'dot': write to file the diagram in dot format.
        'auto': choose from the filename:
            '*.er': writes to a file the markup to generate an ER style diagram.
            '.dot': returns the graph in the dot syntax.
            '.md': writes to a file the markup to generate an Mermaid-JS style diagram
            else: return a graph to the format graph
    :param include_tables: lst of str, table names to include, None means include all
    :param include_columns: lst of str, column names to include, None means include all
    :param exclude_tables: lst of str, table names to exclude, None means exclude nothing
    :param exclude_columns: lst of str, field names to exclude, None means exclude nothing
    :param schema: name of the schema
    """
    try:
        tables, relationships = all_to_intermediary(input, schema=schema)
        tables, relationships = filter_resources(
            tables,
            relationships,
            include_tables=include_tables,
            include_columns=include_columns,
            exclude_tables=exclude_tables,
            exclude_columns=exclude_columns,
        )
        intermediary_to_output = get_output_mode(output, mode)
        intermediary_to_output(tables, relationships, output)
    except ImportError as e:
        module_name = e.message.split()[-1]
        print('Please install {0} using "pip install {0}".'.format(module_name))
    except ParsingException as e:
        sys.stderr.write(e.message)


if __name__ == "__main__":
    cli()



"""
This class allow to transform SQLAlchemy metadata to the intermediary syntax.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable

import sqlalchemy as sa
from sqlalchemy import create_engine
from sqlalchemy.exc import CompileError
from sqlalchemy.ext.automap import AutomapBase, automap_base

from .models import Column, Relation, Table

if TYPE_CHECKING:
    from typing_extensions import Protocol

    class DeclarativeBase(Protocol):
        metadata: sa.MetaData


def relation_to_intermediary(fk: sa.ForeignKey) -> Relation:
    """Transform an SQLAlchemy ForeignKey object to it's intermediary representation."""
    return Relation(
        right_col=format_name(fk.parent.table.fullname),
        left_col=format_name(fk._column_tokens[1]),
        right_cardinality="*",
        left_cardinality="?",
    )


def format_type(typ: Any) -> str:
    """Transforms the type into a nice string representation."""
    try:
        return str(typ)
    except CompileError:
        return "Null"


def format_name(name: Any) -> str:
    """Transforms the name into a nice string representation."""
    return str(name)


def column_to_intermediary(
    col: sa.Column, type_formatter: Callable[[Any], str] = format_type
) -> Column:
    """Transform an SQLAlchemy Column object to it's intermediary representation."""
    return Column(
        name=col.name,
        type=type_formatter(col.type),
        is_key=col.primary_key,
        is_null=col.nullable,
    )


def table_to_intermediary(table: sa.Table) -> Table:
    """Transform an SQLAlchemy Table object to it's intermediary representation."""
    table_columns = getattr(table.c, "_colset", getattr(table.c, "_data", {}).values())
    return Table(
        name=table.fullname,
        columns=[column_to_intermediary(col) for col in table_columns],
    )


def metadata_to_intermediary(
    metadata: sa.MetaData,
) -> tuple[list[Table], list[Relation]]:
    """Transforms SQLAlchemy metadata to the intermediary representation."""
    tables = [table_to_intermediary(table) for table in metadata.tables.values()]
    relationships = [
        relation_to_intermediary(fk)
        for table in metadata.tables.values()
        for fk in table.foreign_keys
    ]
    return tables, relationships


def declarative_to_intermediary(
    base: DeclarativeBase,
) -> tuple[list[Table], list[Relation]]:
    """Transform an SQLAlchemy Declarative Base to the intermediary representation."""
    return metadata_to_intermediary(base.metadata)


def name_for_scalar_relationship(
    base: AutomapBase,
    local_cls: Any,
    referred_cls: type[Any],
    constraint: sa.ForeignKeyConstraint,
) -> str:
    """Overriding naming schemes."""
    return referred_cls.__name__.lower() + "_ref"


def database_to_intermediary(
    database_uri: str, schema: str | None = None
) -> tuple[list[Table], list[Relation]]:
    """Introspect from the database (given the database_uri) to create the intermediary representation."""
    Base = automap_base()
    engine = create_engine(database_uri)
    if schema is not None:
        Base.metadata.schema = schema

    # reflect the tables
    Base.prepare(
        engine, reflect=True, name_for_scalar_relationship=name_for_scalar_relationship
    )
    return declarative_to_intermediary(Base)



from __future__ import annotations

import sys
from argparse import Namespace
from typing import Any


# from https://github.com/mitsuhiko/flask/blob/master/scripts/make-release.py L92
def fail(message: str, *args: Any) -> None:
    print("Error:", message % args, file=sys.stderr)
    sys.exit(1)


def check_args(args: Namespace) -> None:
    """Checks that the args are coherent."""
    check_args_has_attributes(args)
    if args.v:
        non_version_attrs = [v for k, v in args.__dict__.items() if k != "v"]
        print("non_version_attrs", non_version_attrs)
        if len([v for v in non_version_attrs if v is not None]) != 0:
            fail("Cannot show the version number with another command.")
        return
    if args.i is None:
        fail("Cannot draw ER diagram of no database.")
    if args.o is None:
        fail("Cannot draw ER diagram with no output file.")


def check_args_has_attributes(args: Namespace) -> None:
    check_args_has_attribute(args, "i")
    check_args_has_attribute(args, "o")
    check_args_has_attribute(args, "include_tables")
    check_args_has_attribute(args, "include_columns")
    check_args_has_attribute(args, "exclude_tables")
    check_args_has_attribute(args, "exclude_columns")
    check_args_has_attribute(args, "s")


def check_args_has_attribute(args: Namespace, name: str) -> None:
    if not hasattr(args, name):
        raise Exception("{} should be set".format(name))



"""
All the constants used in the module.
"""
from __future__ import annotations

TABLE = (
    '"{}" [label=<<FONT FACE="Helvetica"><TABLE BORDER="0" CELLBORDER="1"'
    ' CELLPADDING="4" CELLSPACING="0">{}{}</TABLE></FONT>>];'
)

START_CELL = '<TR><TD ALIGN="LEFT"><FONT>'
FONT_TAGS = "<FONT>{}</FONT>"
# Used for each row in the table.
ROW_TAGS = "<TR><TD{}>{}</TD></TR>"
GRAPH_BEGINNING = (
    " graph {\n"
    "    graph [rankdir=LR];\n"
    '    node [label="\\N",\n'
    "        shape=plaintext\n"
    "    ];\n"
    "    edge [color=gray50,\n"
    "        minlen=2,\n"
    "        style=dashed\n"
    "    ];\n"
)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['eralchemy2']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy>=1.3.19', 'pygraphviz>=1.7,<2.0,!=1.8']

extras_require = \
{':python_version < "3.8"': ['importlib_metadata>=4.0.0'],
 'ci': ['Flask-SQLAlchemy>=2.5.1,<3.0.0', 'psycopg2>=2.9.3,<3.0.0']}

entry_points = \
{'console_scripts': ['eralchemy2 = eralchemy2.main:cli']}

setup_kwargs = {
    'name': 'eralchemy2',
    'version': '1.3.6',
    'description': 'Simple entity relation (ER) diagrams generation',
    'long_description': '\n[![PyPI Version](https://img.shields.io/pypi/v/eralchemy2.svg)](\nhttps://pypi.org/project/eralchemy2/)\n[![PyPI Downloads](https://img.shields.io/pypi/dm/eralchemy2.svg)](\nhttps://pypi.org/project/eralchemy2/)\n![Github Actions](https://github.com/maurerle/eralchemy2/actions/workflows/python-app.yml/badge.svg)\n\n\n# Entity relation diagrams generator\n\neralchemy2 generates Entity Relation (ER) diagram (like the one below) from databases or from SQLAlchemy models.\nWorks with SQLAlchemy < 1.4 but also with versions greater than 1.4\n\n## Example\n\n![Example for a graph](https://raw.githubusercontent.com/maurerle/eralchemy2/main/newsmeme.svg?raw=true "Example for NewsMeme")\n\n[Example for NewsMeme](https://bitbucket.org/danjac/newsmeme)\n\n## Quick Start\n\n### Install\nTo install eralchemy2, just do:\n\n    $ pip install eralchemy2\n\n`eralchemy2` requires [GraphViz](http://www.graphviz.org/download) to generate the graphs and Python. Both are available for Windows, Mac and Linux.\n\n### Usage from Command Line\n\n#### From a database\n\n    $ eralchemy2 -i sqlite:///relative/path/to/db.db -o erd_from_sqlite.pdf\n\nThe database is specified as a [SQLAlchemy](http://docs.sqlalchemy.org/en/rel_1_0/core/engines.html#database-urls)\ndatabase url.\n\n#### From a markdown file.\n\n    $ curl \'https://raw.githubusercontent.com/maurerle/eralchemy2/main/example/newsmeme.er\' > markdown_file.er\n    $ eralchemy2 -i \'markdown_file.er\' -o erd_from_markdown_file.pdf\n\n#### From a Postgresql DB to a markdown file excluding tables named `temp` and `audit`\n\n    $ eralchemy2 -i \'postgresql+psycopg2://username:password@hostname:5432/databasename\' -o filtered.er --exclude-tables temp audit\n\n#### From a Postgresql DB to a markdown file excluding columns named `created_at` and `updated_at` from all tables\n\n    $ eralchemy2 -i \'postgresql+psycopg2://username:password@hostname:5432/databasename\' -o filtered.er --exclude-columns created_at updated_at\n\n#### From a Postgresql DB to a markdown file for the schema `schema`\n\n    $ eralchemy2 -i \'postgresql+psycopg2://username:password@hostname:5432/databasename\' -s schema\n\n### Usage from Python\n\n```python\nfrom eralchemy2 import render_er\n## Draw from SQLAlchemy base\nrender_er(Base, \'erd_from_sqlalchemy.png\')\n\n## Draw from database\nrender_er("sqlite:///relative/path/to/db.db", \'erd_from_sqlite.png\')\n```\n\n## Architecture\n![Architecture schema](https://raw.githubusercontent.com/maurerle/eralchemy2/main/eralchemy_architecture.png?raw=true "Architecture schema")\n\nThanks to it\'s modular architecture, it can be connected to other ORMs/ODMs/OGMs/O*Ms.\n\n## Contribute\n\nEvery feedback is welcome on the [GitHub issues](https://github.com/maurerle/eralchemy2/issues).\n\nTo run the tests, use : `$ py.test` or `$ tox`.\nSome tests require a local postgres database with a schema named test in a database\nnamed test all owned by a user named eralchemy with a password of eralchemy.\nThis can be easily set up using docker-compose with: `docker-compose up -d`.\n\nAll tested PR are welcome.\n\n## Notes\n\neralchemy2 is a fork of its predecessor [ERAlchemy](https://github.com/Alexis-benoist/eralchemy) by @Alexis-benoist, which is not maintained anymore and does not work with SQLAlchemy > 1.4.\nIf it is maintained again, I\'d like to push the integrated changes upstream.\n\nERAlchemy was inspired by [erd](https://github.com/BurntSushi/erd), though it is able to render the ER diagram directly\nfrom the database and not just only from the `ER` markup language.\n\nReleased under an Apache License 2.0\n\nInitial Creator: Alexis Benoist [Alexis_Benoist](https://twitter.com/Alexis_Benoist)\n',
    'author': 'Florian Maurer',
    'author_email': 'fmaurer@disroot.org',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/maurerle/eralchemy2',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from typing import ClassVar, Iterator

from .models import Column, Drawable, Relation, Table

TYPES: list[type[Drawable]] = [Table, Relation, Column]


class ParsingException(Exception):
    base_traceback = "Error on line {line_nb}: {line}\n{error}"
    hint: ClassVar[str | None] = None

    @property
    def traceback(self) -> str:
        rv = self.base_traceback.format(
            line_nb=getattr(self, "line_nb", "?"),
            line=getattr(self, "line", ""),
            error=self.args[0],
        )
        if self.hint is not None:
            rv += "\nHINT: {}".format(self.hint)
        return rv


class DuplicateTableException(ParsingException):
    pass


class DuplicateColumnException(ParsingException):
    pass


class RelationNoColException(ParsingException):
    hint = "Try to declare the tables before the relationships."


class NoCurrentTableException(ParsingException):
    hint = "Try to declare the tables before the relationships and columns."


def remove_comments_from_line(line: str) -> str:
    if "#" not in line:
        return line.strip()
    return line[: line.index("#")].strip()


def filter_lines_from_comments(lines: list[str]) -> Iterator[tuple[int, str, str]]:
    """Filter the lines from comments and non code lines."""
    for line_nb, raw_line in enumerate(lines):
        clean_line = remove_comments_from_line(raw_line)
        if clean_line == "":
            continue
        yield line_nb, clean_line, raw_line


def parse_line(line: str) -> Drawable:
    for typ in TYPES:
        match = typ.RE.match(line)
        if match:
            return typ.make_from_match(match)
    msg = 'Line "{}" could not be parsed to an object.'
    raise ValueError(msg.format(line))


def _check_no_current_table(new_obj: Drawable, current_table: Table | None) -> None:
    """Raises exception if we try to add a relation or a column
    with no current table."""
    if current_table is None:
        msg = "Cannot add {} before adding table"
        if isinstance(new_obj, Relation):
            raise NoCurrentTableException(msg.format("relation"))
        if isinstance(new_obj, Column):
            raise NoCurrentTableException(msg.format("column"))


def _update_check_inputs(
    current_table: Table | None, tables: list[Table], relations: list[Relation]
) -> None:
    assert current_table is None or isinstance(current_table, Table)
    assert isinstance(tables, list)
    assert all(isinstance(t, Table) for t in tables)
    assert all(isinstance(r, Relation) for r in relations)
    assert current_table is None or current_table in tables


def _check_colname_in_lst(column_name: str, columns_names: list[str]) -> None:
    if column_name not in columns_names:
        msg = 'Cannot add a relation with column "{}" which is undefined'
        raise RelationNoColException(msg.format(column_name))


def _check_not_creating_duplicates(
    new_name: str, names: list[str], type: str, exc: type[Exception]
) -> None:
    if new_name in names:
        msg = 'Cannot add {} named "{}" which is ' "already present in the schema."
        raise exc(msg.format(type, new_name))


def update_models(
    new_obj, current_table: Table | None, tables: list[Table], relations: list[Relation]
) -> tuple[Table | None, list[Table], list[Relation]]:
    """Update the state of the parsing."""
    _update_check_inputs(current_table, tables, relations)
    _check_no_current_table(new_obj, current_table)

    if isinstance(new_obj, Table):
        tables_names = [t.name for t in tables]
        _check_not_creating_duplicates(
            new_obj.name, tables_names, "table", DuplicateTableException
        )
        return new_obj, tables + [new_obj], relations

    if isinstance(new_obj, Relation):
        tables_names = [t.name for t in tables]
        _check_colname_in_lst(new_obj.right_col, tables_names)
        _check_colname_in_lst(new_obj.left_col, tables_names)
        return current_table, tables, relations + [new_obj]

    if isinstance(new_obj, Column):
        assert current_table
        columns_names = [c.name for c in current_table.columns]
        _check_not_creating_duplicates(
            new_obj.name, columns_names, "column", DuplicateColumnException
        )
        current_table.columns.append(new_obj)
        return current_table, tables, relations

    msg = "new_obj cannot be of type {}"
    raise ValueError(msg.format(new_obj.__class__.__name__))


def markdown_file_to_intermediary(filename: str) -> tuple[list[Table], list[Relation]]:
    """Parse a file and return to intermediary syntax."""
    with open(filename) as f:
        lines = f.readlines()
    return line_iterator_to_intermediary(lines)


def line_iterator_to_intermediary(
    line_iterator: list[str],
) -> tuple[list[Table], list[Relation]]:
    """Parse an iterator of str (one string per line) to the intermediary syntax"""
    current_table = None
    tables: list[Table] = []
    relations: list[Relation] = []
    errors: list[ParsingException] = []
    for line_nb, line, raw_line in filter_lines_from_comments(line_iterator):
        try:
            new_obj = parse_line(line)
            current_table, tables, relations = update_models(
                new_obj, current_table, tables, relations
            )
        except ParsingException as e:
            e.line_nb = line_nb  # type:ignore
            e.line = raw_line  # type:ignore
            errors.append(e)
    if len(errors) != 0:
        msg = "eralchemy2 couldn't complete the generation due the {} following errors".format(
            len(errors)
        )
        raise ParsingException(msg + "\n\n".join(e.traceback for e in errors))
    return tables, relations
